/*
 * ALSA SiI9022a  HDMI audio driver
 *
 *  This driver is used by TI81XX devices to config SiI9022a HDMI audio
 *  where no codec is needed.  This file provides stub codec that can be used
 *  in these configurations.
 *  Based on OMAP4 HDMI audio codec driver
 *
 * Author:      ChengYeh Yu <cyyu@synology.com>
 * Copyright:   (C) 2012 synology Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/slab.h>
#include <sound/soc.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/initval.h>
#include <linux/i2c.h>

#define TPI_SAMPLERATE_REG (0x24)
#define TPI_WORDLENGTH_REG (0x25)

static struct snd_soc_codec_driver soc_codec_sii9022a;
static struct i2c_client *gSii9022aI2CClient;

/* null function */
static int sii9022a_set_dai_sysclk(struct snd_soc_dai *codec_dai,
				int clk_id, unsigned int freq, int dir)
{
	return 0;
}

static int sii9022a_set_dai_fmt(struct snd_soc_dai *codec_dai,
			     unsigned int fmt)
{
	return 0;
}

static int sii9022a_hw_params(struct snd_pcm_substream *substream,
			   struct snd_pcm_hw_params *params,
			   struct snd_soc_dai *dai)
{
	u8 data;

	/* select data word length */
	data = i2c_smbus_read_byte_data(gSii9022aI2CClient, TPI_WORDLENGTH_REG);
	data &= ~0xF;

	switch (params_format(params)) {
	case SNDRV_PCM_FORMAT_S16_LE:
		data |= 0x02;
		break;
	case SNDRV_PCM_FORMAT_S20_3LE:
		data |= 0x03;
		break;
	case SNDRV_PCM_FORMAT_S24_LE:
		data |= 0x0B;
		break;
	}
	if (0 >	i2c_smbus_write_byte_data(gSii9022aI2CClient, TPI_WORDLENGTH_REG, data)) {
		printk(KERN_ERR "Cannot program TPI reg at 0x%x",TPI_WORDLENGTH_REG);
	}

	data = i2c_smbus_read_byte_data(gSii9022aI2CClient, TPI_SAMPLERATE_REG);
	data &= ~0xF;
	switch (params->rate_num) {
		case 32000:
			data |= 0x03;
			break;
		case 44100:
			data |= 0x00;
			break;
		case 48000:
			data |= 0x02;
			break;
		case 96000:
			data |= 0x0A;
			break;
		case 192000:
			data |= 0x0E;
			break;
	}

	if (0 >	i2c_smbus_write_byte_data(gSii9022aI2CClient, TPI_SAMPLERATE_REG, data)) {
		printk(KERN_ERR "Cannot program TPI reg at 0x%x",TPI_SAMPLERATE_REG);
	}

	return 0;
}
static struct snd_soc_dai_ops sii9022a_dai_ops = {
	.hw_params  = sii9022a_hw_params,
	.set_sysclk	= sii9022a_set_dai_sysclk,
	.set_fmt	= sii9022a_set_dai_fmt,
};

/* Dummy dai driver for HDMI */
static struct snd_soc_dai_driver sii9022a_dai = {
	.name = "sii9022a",
	.playback = {
		.stream_name = "Playback",
		.channels_min = 2,
		.channels_max = 2,
		.rates = SNDRV_PCM_RATE_48000
			| SNDRV_PCM_RATE_32000
			| SNDRV_PCM_RATE_44100
			| SNDRV_PCM_RATE_96000
			| SNDRV_PCM_RATE_192000 ,
		.formats = SNDRV_PCM_FMTBIT_S16_LE
			| SNDRV_PCM_FMTBIT_S32_LE
			| SNDRV_PCM_FMTBIT_S24_LE,
	},
	.ops = &sii9022a_dai_ops,
};

static int sii9022a_audio_codec_probe(struct platform_device *pdev)
{
	int ret;

	ret = snd_soc_register_codec(&pdev->dev, &soc_codec_sii9022a,
					&sii9022a_dai, 1);
	if (ret < 0)
		printk(KERN_INFO "SiI9022a HDMI Codec register Failed\n");
	else
		printk(KERN_INFO "Registered SiI9022a HDMI audio codec\n");

	return ret;
}

static int sii9022a_audio_codec_remove(struct platform_device *pdev)
{
	snd_soc_unregister_codec(&pdev->dev);
	return 0;
}

static struct platform_driver sii9022a_audio_codec_driver = {
	.probe		= sii9022a_audio_codec_probe,
	.remove		= sii9022a_audio_codec_remove,
	.driver		= {
		.name	= "sii9022a-audio",
		.owner	= THIS_MODULE,
	},
};

static int __devinit sii9022a_audio_i2c_probe(struct i2c_client *client,
		const struct i2c_device_id *id)
{
	int ret = -1;

	if (!i2c_check_functionality(client->adapter,
				I2C_FUNC_SMBUS_BYTE | I2C_FUNC_I2C)) {
		printk(KERN_ERR "No device for 9022a i2c");
		return -ENODEV;
	}

	gSii9022aI2CClient = kzalloc(sizeof(struct i2c_client), GFP_KERNEL);
	if (!gSii9022aI2CClient) {
		printk(KERN_ERR "Allocate fail for 9022a i2c client");
		ret = -ENOMEM;
		goto alloc_failed;
	}

	gSii9022aI2CClient = client;

	ret = 0;
alloc_failed:
	return ret;
}

static int __devexit sii9022a_audio_i2c_remove(struct i2c_client *client)
{
	kfree(gSii9022aI2CClient);
	return 0;
}

static const struct i2c_device_id sii9022a_audio_id[] = {
	{ "sii9022a", 0 },
	{},
};
MODULE_DEVICE_TABLE(i2c, sii9022a_audio_id);

static struct i2c_driver sii9022a_audio_i2c_driver = {
	.driver = {
		   .owner = THIS_MODULE,
		   .name = "sii9022a",
		   },
	.probe = sii9022a_audio_i2c_probe,
	.remove = sii9022a_audio_i2c_remove,
	.id_table = sii9022a_audio_id,
};

static int __init sii9022a_audio_init(void)
{
	int ret = -1;
	if (0 > i2c_add_driver (&sii9022a_audio_i2c_driver)) {
		printk (KERN_ERR "Adding the i2c driver for sii9022a failed");
		goto END;
	}
	if ( 0 > platform_driver_register(&sii9022a_audio_codec_driver)) {
		printk (KERN_ERR "Register sii9022a codec failed");
		goto END;
	}

	ret = 0;
END:
	return ret;
}

static void __exit sii9022a_audio_exit(void)
{
	i2c_del_driver(&sii9022a_audio_i2c_driver);
	platform_driver_unregister(&sii9022a_audio_codec_driver);
}

module_init(sii9022a_audio_init);
module_exit(sii9022a_audio_exit);

MODULE_DESCRIPTION(" SiI9022a Audio Dummy codec Interface");
MODULE_LICENSE("GPL");
